
#define BSD43

#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>		/* struct timeval */
#include <sys/stat.h>    	/* stucture returned by stat */

#ifdef SYSV
#include <sys/times.h>
#include <sys/param.h>
#else
#include <sys/resource.h>
#include <sys/file.h>
#endif

#include <stdio.h>
#include <mpproto.h>
#include <stdlib.h> 	/* for div function */

unsigned int buflen = 1024;		/* length of buffer */
unsigned int blklen = 1024;              /* lnegth of block */
char *buf;			/* ptr to dynamic buffer */
unsigned int nbuf = 1024;		/* number of buffers to send in sinkmode */
int sinkmode;			/* 0=normal I/O, !0=sink/source mode */

extern int errno;

char Usage[] = "\
Usage: tmpl [-options] <in\n\
	-l##	length of a block written to network (default 1024)\n\
	-s	source a pattern to network\n\
	-n##	number of bufs written to network (-s only, default 1024)\n\
	-m##    total number of phases on each node (2<= m <= 32, default 2), must be specified\n\
	-k##    total number of physical process nodes (hostnames) to be used (default 1, >=1),must be specified\n\
	-d      turn debug mode on (default off)\n\
	Note: the relation between block size (l), buffer size, number of buffers (n), and \n\
	      and number of hosts (k) is as follows: \n\
		buffer size (buflen) = l * k \n\
		total size of data (-s only) written to network = n * buflen \n\ 
";

char stats[128];
double t;			/* transmission time */
long double nbytes;			/* bytes on net */

void prep_timer();
double read_timer();
double cput, realt;		/* user, real time (seconds) */

int nrow=2,ncol=1; 		/* nrow -- # of phases; ncol -- # hosts */
void tid2mesh();
int mesh2tid();
static int debug = 0;

main(argc,argv)
int argc;
char **argv;
{

int	numtasks, 		/* total number of MPL tasks in partitiion */
	taskid,			/* task identifier */
	dest,			/* destination task id to send message */
       	arraymsg, 		/* setting a message type */
	source,			/* origin task id of message */
	prevrow,prevcol,myrow,mycol,nextrow,nextcol; /* task id in the 2d mesh */

int icol,msgid,i,total,nread;
int mbuf[4], dontcare, allgrp;
size_t cnt;
char *bufptr;   /* buffer pointer */
int *destination;	/* destination id array */
char *infile,*outfile; /* input/output file name */
int tasks[32],label=1,gid; 
int fd;

	mpc_environ(&numtasks, &taskid);

	if (argc < 3) goto usage; 

	argv++; argc--;
	while( argc>0 && argv[0][0] == '-' )  {
		switch (argv[0][1]) {

		case 'n':
			nbuf = atoi(&argv[0][2]);
			break;
		case 'l':
			blklen = atoi(&argv[0][2]);
			break;
		case 's':
			sinkmode = 1;	/* source or sink */
			break;
		case 'k':
			ncol = atoi(&argv[0][2]);
			break;
		case 'm':
			nrow = atoi(&argv[0][2]);
			break;
		case 'd':
			debug = 1;
			break;
		default:
			goto usage;
		}
		argv++; argc--;
	}

	if ((numtasks != nrow*ncol) || nrow < 2) err("wrong number of tasks");
        mpc_task_query (mbuf, 4, 3);
        dontcare = mbuf[0];
	allgrp = mbuf[3];

	tid2mesh(&myrow,&mycol,taskid);
	nextrow = myrow + 1;
	nextcol = mycol + 1;
	prevrow = myrow - 1;
	prevcol = mycol - 1;

	buflen = blklen * ncol;	
	
	if ( (buflen !=0 && ((buf = (char *)malloc(buflen)) == (char *)NULL)) || 
		((destination = (int *)malloc(ncol*sizeof(int))) == (int *)NULL) )
		err("malloc");

	/* non-sinkmode, assuming a single input/output file for all nodes */
	if (!sinkmode) {
		struct stat stbuf;
        	if ( (infile = getenv("INFILE")) == (char *)NULL ) 
               		err("variable INFILE is not set");
		if ((myrow == 1) && ((fd = open(infile,O_RDONLY)) < 0)) 
			err("open input file");
		if (myrow == nrow) {
        		if ( (outfile = getenv("OUTFILE")) == (char *)NULL ) {
				if ((outfile=(char *)malloc(strlen(infile)+5)) == (char *)NULL) 
					err("malloc");
				strcpy(outfile,infile);
				strcat(outfile,".out");
				if (mycol==ncol) fprintf(stderr,"Input file: %s; Output file: %s\n",infile,outfile);
			}
			if ((fd = creat(outfile,0660)) < 0) 
				err("open output file");
		}
		stat(infile,&stbuf);
		if (buflen != 0) nbuf = (int) (stbuf.st_size/buflen); 
		else nbuf = 1; /* measue latency */
	}
	else pattern( (bufptr=buf), buflen);

	srand(taskid); /* initialize the random generator */
	for (i=0; i<ncol; i++) destination[i] = i+1;

	/* create a new group for final timer synchronization */
	if (myrow == 1 || myrow == nrow) {
		for (i=1; i<=ncol; i++) {
			tasks[i-1] = mesh2tid(1,i);
			tasks[ncol+i-1] = mesh2tid(nrow,i);
		}
		if (mpc_group(2*ncol,tasks,label,&gid)) err("mpc_group");
	}	
		
	errno = 0;
	if (mpc_sync(allgrp)) err("mpc_sync");

	for (i=1; i<=nbuf; i++) {      
		if (myrow != 1) { /* intermediate or final phase, receiver and sender, or receiver only */
			if (debug) {
				fprintf(stderr,"**%d**: (%d,*) --> [%d,%d] receiving ... \n",
					i,prevrow,myrow,mycol);	
				total = 0;
			}
			bufptr = buf;	
			for (icol=0; icol<ncol; icol++) {	
				source = dontcare;
				arraymsg = i;
				/* if strict phase-level synchronization is not required, uncommnet
				   the line below which will increase overall bandwidth */
				/* arraymsg = dontcare; */ 
				
				if (mpc_brecv(bufptr,blklen,&source,&arraymsg,&cnt)||cnt != blklen)
					err("error in mpc_recv");
				bufptr += blklen;
				if (debug) {
					int srow,scol;
					tid2mesh(&srow,&scol,source);
					total += scol;
					fprintf(stderr,"**%d**: (%d,%d) --> [%d,%d] done.\n",
						 i,srow,scol,myrow,mycol); 
				}
			}
			if (debug && total != ((ncol+1)*ncol/2)) {
				fprintf(stderr,"###%d###: Warning - (%d,%d) got more than 1 block from a same node (%d != %d) ###\n",
					i,myrow,mycol,total,((ncol+1)*ncol/2));	
		     			err("wrong order of send/receive");
			}
			if ( (myrow == nrow) && !sinkmode) {  /* final phase, receiver only */
				if (write(fd,buf,buflen) != buflen) err("write I/O");
			}		
		}

		if (myrow != nrow) { /* first or intermediate phase, send to next phase */
			if (myrow == 1) {													
				if (i == 1) { /* do anything needs to be done initially */
					if (new_sleep(10)) err("new_sleep"); /* sleep for a while until all receiving tasks are ready */
					if (mycol==1) fprintf(stderr,"starting ..\n");
					prep_timer(); /* start the timer */
				}

				/* if ( !sinkmode && ((nread=read(fd,buf,buflen)) <= 0)) {
						nbuf = i - 1;
						break;
				} */
				if (sinkmode)
					pattern( (bufptr=buf), buflen); 
				else if ( (nread=read(fd,buf,buflen)) <= 0) {
						nbuf = i - 1;
						break;
				}
			}	
			if (debug) fprintf(stderr,"**%d**: [%d,%d] --> (%d,*) sending ...\n",
				 	i,myrow,mycol,nextrow);		
			for (icol=0; icol<ncol; icol++) { /* randomize the destination hosts */
				register int itmp,j;
				j = rand()%ncol;
				itmp = destination[icol]; /* swap the two hosts */
				destination[icol] = destination[j];
				destination[j] = itmp;
			}
			bufptr = buf;	
			for (icol=0; icol<ncol; icol++) {  /* needs to randomly choose destination node */
				dest = mesh2tid(nextrow,destination[icol]);
				arraymsg = i; /* this is required to strict phase-level synchronization, */
					      /* i.e., only after all k blocks received, each block from */
					      /* a different process A on one of the k hosts,		 */
			   		      /* an intermediate process B should send them to C. 	 */
			  		      /* using msg type to make sure next row finished receiving */
					      /* all blocksfrom k node in the previous run (indexed by i)*/
				if (mpc_bsend(bufptr,blklen,dest,arraymsg)) err("mpc_bsend");
				bufptr += blklen;
			}
		}
	}			
	if (myrow == 1 || myrow == nrow) {
		if (!sinkmode) close(fd);
 		if (mpc_sync(gid)) err("mpc_sync final"); /* make sure all 2*ncol initial and final finished */			
	}
	if (myrow == 1) {
		(void)read_timer(stats,sizeof(stats));
		if (mycol==1) fprintf(stderr,"tmpl: nbuf=%d, buflen=%d, blklen=%d, nphase=%d, nhosts=%d\n",
					nbuf, buflen, blklen, nrow, ncol);
		nbytes = (long double) ((nbuf*buflen)*(ncol))*ncol*(nrow-1)/(1024.*1024.);  /* note B -> C on the same node is included */
		if( cput <= 0.0 )  cput = 0.001;
		if( realt <= 0.0 )  realt = 0.001;
		fprintf(stderr,"(%d,%d): %f MB processed\n",myrow,mycol, (long double)nbytes);
		fprintf(stderr,"(%d,%d): %9g CPU sec  = %9g MB/cpu sec,  %9g Mbits/cpu sec\n",myrow,mycol,
			cput,
			((long double)nbytes)/cput,
			((long double)nbytes)*8/cput );
		fprintf(stderr,"(%d,%d): %9g real sec = %9g MB/real sec, %9g Mbits/sec\n",myrow,mycol,
			realt,
			((long double)nbytes)/realt,
			((long double)nbytes)*8/realt );
	}
	exit(0);

usage:
	if (taskid == 0) fprintf(stderr,Usage);
	exit(1);
}

static int mesh2tid (row,col)
int row,col;
{

	/* if ((row <= 0) || (row >nrow) || (col <= 0) || (col >ncol))
		err("wrong row or col id"); */
	return ((row-1)*ncol+col-1);
}

static void tid2mesh(row,col,tid)
int *row,*col,tid;
{
	struct div_t tmp;

	/*if (tid >= nrow*ncol)
		err("wrong tid");*/
	tmp = div(tid,ncol);
	*row = tmp.quot+1;
	*col = tmp.rem+1;
}
err(s)
char *s;
{
	fprintf(stderr,"error in %s, errno=%d\n",s,errno);
	mpc_stopall(errno);
	exit(1);
}

pattern( cp, cnt )
register char *cp;
register int cnt;
{
	register char c;
	c = 0;
	while( cnt-- > 0 )  {
		/* while( !isprint((c&0x7F)) )  c++;
		*cp++ = (c++&0x7F);*/
		*cp++ = 0;
	}
}

/******* timing *********/

#ifdef SYSV
extern long time();
static long time0;
static struct tms tms0;
#else
static struct	timeval time0;	/* Time at which timeing started */
static struct	rusage ru0;	/* Resource utilization at the start */

static void prusage();
static void tvadd();
static void tvsub();
static void psecs();
#endif

/*
 *			P R E P _ T I M E R
 */
void
prep_timer()
{
#ifdef SYSV
	(void)time(&time0);
	(void)times(&tms0);
#else
	gettimeofday(&time0, (struct timezone *)0);
	getrusage(RUSAGE_SELF, &ru0);
#endif
}

/*
 *			R E A D _ T I M E R
 * 
 */
double
read_timer(str,len)
char *str;
{
#ifdef SYSV
	long now;
	struct tms tmsnow;
	char line[132];

	(void)time(&now);
	realt = now-time0;
	(void)times(&tmsnow);
	cput = tmsnow.tms_utime - tms0.tms_utime;
	cput /= HZ;
	if( cput < 0.00001 )  cput = 0.01;
	if( realt < 0.00001 )  realt = cput;
	sprintf(line,"%g CPU secs in %g elapsed secs (%g%%)",
		cput, realt,
		cput/realt*100 );
	(void)strncpy( str, line, len );
	return( cput );
#else
	/* BSD */
	struct timeval timedol;
	struct rusage ru1;
	struct timeval td;
	struct timeval tend, tstart;
	char line[132];

	getrusage(RUSAGE_SELF, &ru1);
	gettimeofday(&timedol, (struct timezone *)0);
	prusage(&ru0, &ru1, &timedol, &time0, line);
	(void)strncpy( str, line, len );

	/* Get real time */
	tvsub( &td, &timedol, &time0 );
	realt = td.tv_sec + ((double)td.tv_usec) / 1000000;

	/* Get CPU time (user+sys) */
	tvadd( &tend, &ru1.ru_utime, &ru1.ru_stime );
	tvadd( &tstart, &ru0.ru_utime, &ru0.ru_stime );
	tvsub( &td, &tend, &tstart );
	cput = td.tv_sec + ((double)td.tv_usec) / 1000000;
	if( cput < 0.00001 )  cput = 0.00001;
	return( cput );
#endif
}

#ifndef SYSV
static void
prusage(r0, r1, e, b, outp)
	register struct rusage *r0, *r1;
	struct timeval *e, *b;
	char *outp;
{
	struct timeval tdiff;
	register time_t t;
	register char *cp;
	register int i;
	int ms;

	t = (r1->ru_utime.tv_sec-r0->ru_utime.tv_sec)*100+
	    (r1->ru_utime.tv_usec-r0->ru_utime.tv_usec)/10000+
	    (r1->ru_stime.tv_sec-r0->ru_stime.tv_sec)*100+
	    (r1->ru_stime.tv_usec-r0->ru_stime.tv_usec)/10000;
	ms =  (e->tv_sec-b->tv_sec)*100 + (e->tv_usec-b->tv_usec)/10000;

#define END(x)	{while(*x) x++;}
	cp = "%Uuser %Ssys %Ereal %P %Xi+%Dd %Mmaxrss %F+%Rpf %Ccsw";
	for (; *cp; cp++)  {
		if (*cp != '%')
			*outp++ = *cp;
		else if (cp[1]) switch(*++cp) {

		case 'U':
			tvsub(&tdiff, &r1->ru_utime, &r0->ru_utime);
			sprintf(outp,"%d.%01d", tdiff.tv_sec, tdiff.tv_usec/100000);
			END(outp);
			break;

		case 'S':
			tvsub(&tdiff, &r1->ru_stime, &r0->ru_stime);
			sprintf(outp,"%d.%01d", tdiff.tv_sec, tdiff.tv_usec/100000);
			END(outp);
			break;

		case 'E':
			psecs(ms / 100, outp);
			END(outp);
			break;

		case 'P':
			sprintf(outp,"%d%%", (int) (t*100 / ((ms ? ms : 1))));
			END(outp);
			break;

		case 'W':
			i = r1->ru_nswap - r0->ru_nswap;
			sprintf(outp,"%d", i);
			END(outp);
			break;

		case 'X':
			sprintf(outp,"%d", t == 0 ? 0 : (r1->ru_ixrss-r0->ru_ixrss)/t);
			END(outp);
			break;

		case 'D':
			sprintf(outp,"%d", t == 0 ? 0 :
			    (r1->ru_idrss+r1->ru_isrss-(r0->ru_idrss+r0->ru_isrss))/t);
			END(outp);
			break;

		case 'K':
			sprintf(outp,"%d", t == 0 ? 0 :
			    ((r1->ru_ixrss+r1->ru_isrss+r1->ru_idrss) -
			    (r0->ru_ixrss+r0->ru_idrss+r0->ru_isrss))/t);
			END(outp);
			break;

		case 'M':
			sprintf(outp,"%d", r1->ru_maxrss/2);
			END(outp);
			break;

		case 'F':
			sprintf(outp,"%d", r1->ru_majflt-r0->ru_majflt);
			END(outp);
			break;

		case 'R':
			sprintf(outp,"%d", r1->ru_minflt-r0->ru_minflt);
			END(outp);
			break;

		case 'I':
			sprintf(outp,"%d", r1->ru_inblock-r0->ru_inblock);
			END(outp);
			break;

		case 'O':
			sprintf(outp,"%d", r1->ru_oublock-r0->ru_oublock);
			END(outp);
			break;
		case 'C':
			sprintf(outp,"%d+%d", r1->ru_nvcsw-r0->ru_nvcsw,
				r1->ru_nivcsw-r0->ru_nivcsw );
			END(outp);
			break;
		}
	}
	*outp = '\0';
}

static void
tvadd(tsum, t0, t1)
	struct timeval *tsum, *t0, *t1;
{

	tsum->tv_sec = t0->tv_sec + t1->tv_sec;
	tsum->tv_usec = t0->tv_usec + t1->tv_usec;
	if (tsum->tv_usec > 1000000)
		tsum->tv_sec++, tsum->tv_usec -= 1000000;
}

static void
tvsub(tdiff, t1, t0)
	struct timeval *tdiff, *t1, *t0;
{

	tdiff->tv_sec = t1->tv_sec - t0->tv_sec;
	tdiff->tv_usec = t1->tv_usec - t0->tv_usec;
	if (tdiff->tv_usec < 0)
		tdiff->tv_sec--, tdiff->tv_usec += 1000000;
}

static void
psecs(l,cp)
long l;
register char *cp;
{
	register int i;

	i = l / 3600;
	if (i) {
		sprintf(cp,"%d:", i);
		END(cp);
		i = l % 3600;
		sprintf(cp,"%d%d", (i/60) / 10, (i/60) % 10);
		END(cp);
	} else {
		i = l;
		sprintf(cp,"%d", i / 60);
		END(cp);
	}
	i %= 60;
	*cp++ = ':';
	sprintf(cp,"%d%d", i / 10, i % 10);
}
#endif

int new_sleep (amount)
int amount;
{
        struct timestruc_t Requested, Remaining;
        double famount = amount;
        int rc;
        while (famount > 0.0) {
                Requested.tv_sec = (int) famount;
                Requested.tv_nsec =
                        (int) ((famount - Requested.tv_sec)*1000000000.);
                rc = nsleep ( &Requested, &Remaining );
                if ((rc == -1) && (errno == EINTR)) {
                        /* Sleep interrupted.  Resume it */
                        famount = Remaining.tv_sec + Remaining.tv_nsec /
                                                        1000000000.;
                        continue;
                }
                else /* Completed sleep.  Set return to zero */
                {
                                return (0);
                }
        }    /* end of while */

        /* famount = 0.; exit */
        return (0);
}


	
	 
